from contextlib import contextmanager
import json
import string
import uuid
from pycassa.cassandra.ttypes import NotFoundException
from agamemnon.graph_constants import RELATIONSHIP_KEY_PATTERN, OUTBOUND_RELATIONSHIP_CF, RELATIONSHIP_INDEX, ENDPOINT_NAME_TEMPLATE, INBOUND_RELATIONSHIP_CF
import pycassa
from agamemnon.cassandra import CassandraDataStore
from agamemnon.memory import InMemoryDataStore
from agamemnon.exceptions import NodeNotFoundException
import agamemnon.primitives as prim


class DataStore(object):
    def __init__(self, delegate):
        self.delegate = delegate

    @contextmanager
    def batch(self):
        self.delegate.start_batch()
        yield
        self.delegate.commit_batch()

    def get(self, type, row_key, super_column_key=None):
        column_family = self.delegate.get_cf(type)
        if column_family is None:
            self.delegate.create_cf(type)
            column_family = self.delegate.get_cf(type)
        if super_column_key is None:
            return self.deserialize_value(column_family.get(row_key))
        else:
            return self.deserialize_value(column_family.get(row_key, super_column=super_column_key))

    def delete(self, type, key, super_key=None):
        if super_key is None:
            self.delegate.get_cf(type).remove(key)
        else:
            self.delegate.get_cf(type).remove(key, super_key)

    def insert(self, type, key, args, super_key=None):
        if not self.delegate.cf_exists(type):
            column_family = self.delegate.create_cf(type)
        else:
            column_family = self.delegate.get_cf(type)
        serialized = self.serialize_columns(args)
        if super_key is None:
            column_family.insert(key, serialized)
        else:
            column_family.insert(key, {super_key: serialized})

    def get_outgoing_relationships(self, source_node, rel_type, count=100):
        source_key = RELATIONSHIP_KEY_PATTERN % (source_node.type, source_node.key)
        cf = self.delegate.get_cf(OUTBOUND_RELATIONSHIP_CF)
        #Ok, this is weird.  So in order to get a column slice, you need to provide a start that is <= your first column
        #id, and a finish which is >= your last column.  Since our columns are sorted by ascii, this means we need to go
        #from rel_type_ to rel_type` because "`" is the char 1 greater than "_", so this will get anything which starts
        #rel_type_.  Now I realize that this could problems when there is a "_" in the relationship name, so we will
        #probably need a different delimiter.
        #TODO: fix delimiter
        try:
            super_columns = cf.get(source_key, column_start='%s__' % rel_type, column_finish='%s_`' % rel_type,
                                   column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_outgoing_relationship(rel_type, source_node, super_column) for super_column in
                super_columns.items()]


    def get_incoming_relationships(self, target_node, rel_type, count=100):
        target_key = RELATIONSHIP_KEY_PATTERN % (target_node.type, target_node.key)
        cf = self.delegate.get_cf(INBOUND_RELATIONSHIP_CF)
        #Ok, this is weird.  So in order to get a column slice, you need to provide a start that is <= your first column
        #id, and a finish which is >= your last column.  Since our columns are sorted by ascii, this means we need to go
        #from rel_type_ to rel_type` because "`" is the char 1 greater than "_", so this will get anything which starts
        #rel_type_.  Now I realize that this could problems when there is a "_" in the relationship name, so we will
        #probably need a different delimiter.
        #TODO: fix delimiter
        try:
            super_columns = cf.get(target_key, column_start='%s__' % rel_type, column_finish='%s_`' % rel_type,
                                   column_count=count)
        except NotFoundException:
            super_columns = {}
        return [self.get_incoming_relationship(rel_type, target_node, super_column) for super_column in
                super_columns.items()]

    def get_outgoing_relationship(self, rel_type, source_node, super_column):
        """
        Process the contents of a SuperColumn to extract the relationship and to_node properties and return
        a constructed relationship
        """
        rel_key = super_column[0]
        target_node_key = None
        target_node_type = None
        target_attributes = {}
        rel_attributes = {}
        for column in super_column[1].keys():
            value = super_column[1][column]
            if column == 'target__type':
                target_node_type = value
            elif column == 'target__key':
                target_node_key = value
            elif column.startswith('target__'):
                target_attributes[column[8:]] = value
            else:
                rel_attributes[column] = value
        return prim.Relationship(rel_key, source_node,
                                 prim.Node(self, target_node_type, target_node_key, target_attributes), self
                                 , rel_type,
                                 rel_attributes)

    def get_incoming_relationship(self, rel_type, target_node, super_column):
        """
        Process the contents of a SuperColumn to extract an incoming relationship and the associated from_node and
        return a constructed relationship
        """
        rel_key = super_column[0]
        source_node_key = None
        source_node_type = None
        source_attributes = {}
        rel_attributes = {}
        for column in super_column[1].keys():
            value = super_column[1][column]
            if column == 'source__type':
                source_node_type = value
            elif column == 'source__key':
                source_node_key = value
            elif column.startswith('source__'):
                source_attributes[column[8:]] = value
            else:
                rel_attributes[column] = value
        return prim.Relationship(rel_key, prim.Node(self, source_node_type, source_node_key, source_attributes),
                                 target_node,
                                 self, rel_type, rel_attributes)


    def delete_relationship(self, rel_type, rel_id, from_type, from_key, to_type, to_key):
        from_key = ENDPOINT_NAME_TEMPLATE % (from_type, from_key)
        to_key = ENDPOINT_NAME_TEMPLATE % (to_type, to_key)

        with self.batch():
            self.delegate.remove(self.delegate.get_cf(INBOUND_RELATIONSHIP_CF), to_key,
                                 super_column=RELATIONSHIP_KEY_PATTERN % (rel_type, rel_id))
            self.delegate.remove(self.delegate.get_cf(OUTBOUND_RELATIONSHIP_CF), from_key,
                                 super_column=RELATIONSHIP_KEY_PATTERN % (rel_type, rel_id))

    def create_relationship(self, rel_type, source_node, target_node, key=None, args=dict()):
        if key is None:
            key = str(uuid.uuid4())
            #node relationship types
        rel_key = RELATIONSHIP_KEY_PATTERN % (rel_type, key)
        with self.batch():
            #outbound_cf
            columns = {'rel_type': rel_type, 'rel_key': key}
            #add relationship attributes
            columns.update(args)
            rel_attr = dict(columns)
            #add target attributes
            columns['target__type'] = target_node.type.encode('ascii')
            columns['target__key'] = target_node.key.encode('ascii')
            target_attributes = target_node.attributes
            for attribute_key in target_attributes.keys():
                columns['target__%s' % attribute_key] = target_attributes[attribute_key]
            columns['source__type'] = source_node.type.encode('ascii')
            columns['source__key'] = source_node.key.encode('ascii')
            source_attributes = source_node.attributes
            for attribute_key in source_attributes.keys():
                columns['source__%s' % attribute_key] = source_attributes[attribute_key]

            source_key = ENDPOINT_NAME_TEMPLATE % (source_node.type, source_node.key)
            target_key = ENDPOINT_NAME_TEMPLATE % (target_node.type, target_node.key)
            inbound_rel_index_cf = self.delegate.get_cf(INBOUND_RELATIONSHIP_CF)
            outbound_rel_index_cf = self.delegate.get_cf(OUTBOUND_RELATIONSHIP_CF)
            serialized = self.serialize_columns(columns)
            self.delegate.insert(outbound_rel_index_cf, source_key, {rel_key: serialized})
            self.delegate.insert(inbound_rel_index_cf, target_key, {rel_key: serialized})

            relationship_index_cf = self.delegate.get_cf(RELATIONSHIP_INDEX)
            # Add entries in the relationship index
            self.delegate.insert(relationship_index_cf, source_key,
                    {target_node.key: {rel_type: '%s__outgoing' % rel_key}})
            self.delegate.insert(relationship_index_cf, target_key,
                    {source_node.key: {rel_type: '%s__incoming' % rel_key}})

        #created relationship object
        return prim.Relationship(rel_key, source_node, target_node, self, rel_type, rel_attr)

    def has_relationship(self, node_a, node_b_key, rel_type):
        """
        This determines if two nodes have a relationship of the specified type.

        > ds = DataStore()
        > node_a =

        """
        index = self.delegate.get_cf(RELATIONSHIP_INDEX)
        node_a_row_key = ENDPOINT_NAME_TEMPLATE % (node_a.type, node_a.key)
        rel_list = []
        try:
            rels = index.get(node_a_row_key, super_column=node_b_key, column_start=rel_type,
                             column_finish='%s`' % rel_type)
            for rel in rels.values():
                if rel.endswith('__incoming'):
                    rel_id = string.replace(rel, '__incoming', '')
                    relationship = self.get_incoming_relationship(rel_type, node_a,
                        (rel_id, self.get(INBOUND_RELATIONSHIP_CF, node_a_row_key, rel_id)))
                elif rel.endswith('__outgoing'):
                    rel_id = string.replace(rel, '__outgoing', '')
                    relationship = self.get_outgoing_relationship(rel_type, node_a,
                        (rel_id, self.get(OUTBOUND_RELATIONSHIP_CF, node_a_row_key, rel_id)))
                else:
                    continue
                rel_list.append(relationship)
        except NotFoundException:
            pass
        return rel_list

    def create_node(self, type, key, args=None, reference=False):
        node = prim.Node(self, type, key, args)
        if args is not None:
            serialized = self.serialize_columns(args)
        else:
            serialized = {}
        self.insert(type, key, serialized)
        if not reference:
            #this adds the created node to the reference node for this type of object
            #that reference node functions as an index to easily access all nodes of a specific type
            reference_node = self.get_reference_node(type)
            reference_node.instance(node, key=key)
        return node

    def delete_node(self, node):
        node_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)
        try:
            outbound_results = self.get(OUTBOUND_RELATIONSHIP_CF, node_key)
        except NotFoundException:
            outbound_results = {}
        try:
            inbound_results = self.get(INBOUND_RELATIONSHIP_CF, node_key)
        except NotFoundException:
            inbound_results = {}

        with self.batch():
            for key in outbound_results.keys():
                rel = outbound_results[key]
                self.delegate.remove(self.delegate.get_cf(INBOUND_RELATIONSHIP_CF),
                                     RELATIONSHIP_KEY_PATTERN % (rel['target__type'], rel['target__key']),
                                     super_column='%s__%s' % (rel['rel_type'], rel['rel_key']))
            for key in inbound_results.keys():
                rel = inbound_results[key]
                self.delegate.remove(self.delegate.get_cf(OUTBOUND_RELATIONSHIP_CF),
                                     RELATIONSHIP_KEY_PATTERN % (rel['source__type'], rel['source__key']),
                                     super_column='%s__%s' % (rel['rel_type'], rel['rel_key']))
            self.delegate.remove(self.delegate.get_cf(OUTBOUND_RELATIONSHIP_CF), node_key)
            self.delegate.remove(self.delegate.get_cf(INBOUND_RELATIONSHIP_CF), node_key)
            self.delegate.remove(self.delegate.get_cf(node.type), node.key)

    def save_node(self, node):
        """
        This needs to update the entry in the type table as well as all of the relationships
        """
        source_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)
        target_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)

        try:
            outbound_results = self.get(OUTBOUND_RELATIONSHIP_CF, source_key)
        except NotFoundException:
            outbound_results = {}
        try:
            inbound_results = self.get(INBOUND_RELATIONSHIP_CF, target_key)
        except NotFoundException:
            inbound_results = {}
        outbound_columns = {'source__type': node.type.encode('utf-8'), 'source__key': node.key.encode('utf-8')}
        node_attributes = node.attributes
        for attribute_key in node.attributes.keys():
            outbound_columns['source__%s' % attribute_key] = node_attributes[attribute_key]
        for key in outbound_results.keys():
            target = outbound_results[key]
            target_key = ENDPOINT_NAME_TEMPLATE % (target['target__type'], target['target__key'])
            target.update(outbound_columns)
            serialized = self.serialize_columns(target)
            self.insert(OUTBOUND_RELATIONSHIP_CF, source_key, serialized, key)
            self.insert(INBOUND_RELATIONSHIP_CF, target_key, serialized, key)
        inbound_columns = {'target__type': node.type.encode('utf-8'), 'target__key': node.key.encode('utf-8')}
        for attribute_key in node.attributes.keys():
            inbound_columns['target__%s' % attribute_key] = node_attributes[attribute_key]
        for key in inbound_results.keys():
            source = inbound_results[key]
            source_key = ENDPOINT_NAME_TEMPLATE % (source['source__type'], source['source__key'])
            target_key = ENDPOINT_NAME_TEMPLATE % (node.type, node.key)
            source.update(inbound_columns)
            serialized = self.serialize_columns(source)
            self.insert(OUTBOUND_RELATIONSHIP_CF, source_key, serialized, key)
            self.insert(INBOUND_RELATIONSHIP_CF, target_key, serialized, key)
        self.insert(node.type, node.key, self.serialize_columns(node.attributes))

    def get_node(self, type, key):
        try:
            values = self.get(type, key)
        except NotFoundException:
            raise NodeNotFoundException()
        return prim.Node(self, type, key, values)

    def get_reference_node(self, name):
        """
        Nodes returned here are very easily referenced by name and then function as an index for all attached nodes
        The most typical use case is to index all of the nodes of a certain type, but the functionality is not limited
        to this.
        """
        try:
            node = self.get_node('reference', name)
        except NodeNotFoundException:
            node = self.create_node('reference', name, {'reference': 'reference'}, reference=True)
        return node

    def deserialize_value(self, value):
        if isinstance(value, dict):
            return self.deserialize_columns(value)
        if not value.startswith('$'):
            return value
        type = value[:2]
        content = value[2:]
        if type == '$b':
            return bool(content)
        elif type == '$i':
            return int(content)
        elif type == '$l':
            return long(content)
        elif type == '$f':
            return float(content)

    def serialize_value(self, value):
        if isinstance(value, bool):
            return '$b%r' % value
        elif isinstance(value, int):
            return '$i%r' % value
        elif isinstance(value, long):
            return '$l%r' % value
        elif isinstance(value, float):
            return '$f%r' % value
        elif isinstance(value, str):
            return value
        elif isinstance(value, dict):
            return self.serialize_columns(value)
        else:
            raise TypeError('Cannot serialize: %s' % type(value))

    def deserialize_columns(self, columns):
        return dict([(key, self.deserialize_value(value)) for key, value in columns.items()])

    def serialize_columns(self, columns):
        return dict([(key, self.serialize_value(value)) for key, value in columns.items()])
        
    def __getattr__(self, item):
        if not item in self.__dict__:
            return getattr(self.delegate, item)


def load_from_settings(settings, prefix='agamemnon.'):
    if settings["%skeyspace" % prefix] == 'memory':
        ds_to_wrap = InMemoryDataStore()
    else:
        ds_to_wrap = CassandraDataStore(settings['%skeyspace' % prefix],
                                        pycassa.connect(settings["%skeyspace" % prefix],
                                                        json.loads(settings["%shost_list" % prefix])),
                                        system_manager=pycassa.system_manager.SystemManager(
                                            json.loads(settings["%shost_list" % prefix])[0]))
    return DataStore(ds_to_wrap)


#from contextlib import nested
#import unittest
#from mock import Mock, patch
#from nose.plugins.attrib import attr
#import pycassa
#from pycassa.cassandra.ttypes import NotFoundException
#import agamemnon
#import agamemnon.graph_constants
#import agamemnon.cassandra
#import agamemnon.factory
#import agamemnon.primitives as graph_prim
#from agamemnon.graph_constants import ENDPOINT_NAME_TEMPLATE, RELATIONSHIP_KEY_PATTERN, RELATIONSHIP_INDEX
#
#relationship_type = 'pig_cow_alliance'
#relationship_types_cf = '__relationship__types'
#type_cf = 'test_type'
#
#class BatchMutatorMock(object):
#    """
#     because we are using the batch.Mutator context manager, we need to implement the magic methods
#     __enter__ and __exit__ as they will be called to provide the contextualized object.
#     contextualized_batch_mock is that object and is used for all of out assertions
#    """
#
#    def __init__(self):
#        self._contextualized_batch_mock = Mock()
#        self.__enter__ = Mock()
#        self.__enter__.return_value = self._contextualized_batch_mock
#        self.__exit__ = Mock()
#
#    @property
#    def contextualized_batch_mock(self):
#        return self._contextualized_batch_mock
#
#
#class CassandraDataStoreTests(unittest.TestCase):
#    def setUp(self):
#        self._pool = Mock()
#        self._system_manager = Mock()
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            def side_effect(self, name):
#                return name
#
#            cf_mock.side_effect = side_effect
#            self.store = agamemnon.factory.DataStore(
#                agamemnon.cassandra.CassandraDataStore('test_space', self._pool, self._system_manager))
#        self.pig = graph_prim.Node(self.store, 'test_type', 'pig')
#        self.cow = graph_prim.Node(self.store, 'test_type', 'cow')
#
#    @attr(module='agamemnon', type='unit')
#    def test_get_non_existent_cf(self):
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            cf_mock.side_effect = NotFoundException()
#            self.failIf(self.store.cf_exists('test_type'))
#
#    @attr(module='agamemnon', type='unit')
#    def test_create_node(self):
#        #create the node
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            def new_cf(*args):
#                if new_cf.first:
#                    new_cf.first = False
#                    raise NotFoundException()
#                else:
#                    return cf_mock.return_value
#
#            new_cf.first = True
#            cf_mock.side_effect = new_cf
#
#            def side_effect(name):
#                if name is 'test_type':
#                    return {'reference': 'reference'}
#                else:
#                    return Mock()
#
#            cf_mock.return_value.get.side_effect = side_effect
#            self.store.create_node('test_type', 'cow', {'sound': 'moo'})
#            self._system_manager.create_column_family.assert_called_with('test_space', 'test_type', super=False,
#                                                                         comparator_type=pycassa.system_manager.ASCII_TYPE)
#            cf_mock.return_value.insert.assert_called_with('cow', {'sound': 'moo'})
#
#    @attr(module='agamemnon', type='unit')
#    def test_create_node_existing_cf(self):
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            def side_effect(name):
#                if name is 'test_type':
#                    return {'reference': 'reference'}
#                else:
#                    return Mock()
#
#            cf_mock.return_value.get.side_effect = side_effect
#            self.store.create_node('test_type', 'pig', {'sound': 'oink'})
#            self.failIf(self._system_manager.create_column_family.called)
#
#    @attr(module='agamemnon', type='unit')
#    def test_create_duplicate_node(self):
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            def side_effect(name):
#                if name is 'test_type':
#                    return {'reference': 'reference'}
#                else:
#                    return Mock()
#
#            cf_mock.return_value.get.side_effect = side_effect
#            self.store.create_node('test_type', 'pig', {'sound': 'oink'})
#            cf_mock.return_value.insert.assert_called_with('pig', {'sound': 'oink'})
#
#
#    @attr(module='agamemnon', type='unit')
#    def test_get_node(self):
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            cf_mock.return_value.get.return_value = {'sound': 'oink'}
#            pig = self.store.get_node('test_type', 'pig')
#            cf_mock.return_value.get.assert_called_with('pig')
#            self.failUnlessEqual(pig.type, 'test_type')
#            self.failUnlessEqual(pig.key, 'pig')
#            self.failUnlessEqual(pig['sound'], 'oink')
#
#    @attr(module='agamemnon', type='unit')
#    def test_get_non_existent_node(self):
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            #Test if node is not in datastore
#            cf_mock.return_value.get.side_effect = NotFoundException()
#            try:
#                self.store.get_node('test_type', 'pig')
#                self.fail()
#            except agamemnon.NodeNotFoundException:
#                pass
#
#    @attr(module='agamemnon', type='unit')
#    def test_delete_node_simple(self):
#        node_mock = Mock()
#        node_mock.remove = Mock()
#        out_bound_rel_mock = Mock()
#        out_bound_rel_mock.get.return_value = {}
#        in_bound_rel_mock = Mock()
#        in_bound_rel_mock.get.return_value = {}
#
#        def side_effect(pool, type):
#            if type == 'outbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX:
#                return out_bound_rel_mock
#            elif type == 'inbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX:
#                return in_bound_rel_mock
#            elif type == type_cf:
#                return node_mock
#            else:
#                raise Exception('unknown cf: %s' % type)
#
#        with nested(patch('pycassa.columnfamily.ColumnFamily'),
#                    patch('pycassa.batch.Mutator')) as (cf_mock, batch_mock):
#            batch_mock.return_value = BatchMutatorMock()
#            contextualized_batch_mock = batch_mock.return_value.contextualized_batch_mock
#            cf_mock.side_effect = side_effect
#            pig = graph_prim.Node(self.store, 'test_type', 'pig')
#            pig.delete()
#            contextualized_batch_mock.remove.assert_called_with(node_mock, 'pig')
#
#
#    @attr(module='agamemnon', type='unit')
#    def test_create_relationship(self):
#        # These are the inserts we expect to get.  The first one is adding friend to the list of relation types
#        # the second and third are the to and from relationship inserts.  This is a pretty complicated action.  It
#        # executes 7 inserts to create a relationship.  This could be cut down by eliminating the type specific
#        # column families, but it is not clear that would make a significant difference in performance
#        expected_calls = [
#            (('outbound__%s' % RELATIONSHIP_INDEX, ENDPOINT_NAME_TEMPLATE % ('test_type', 'pig'),
#                  {RELATIONSHIP_KEY_PATTERN % ('friend', 'pig_cow_alliance'): {'rel_type': 'friend',
#                                                                               'rel_key': 'pig_cow_alliance',
#                                                                               'source__type': 'test_type',
#                                                                               'source__key': 'pig',
#                                                                               'source__sound': 'oink',
#                                                                               'target__type': 'test_type',
#                                                                               'target__key': 'cow',
#                                                                               'target__sound': 'moo'}}), {}),
#
#            (('inbound__%s' % RELATIONSHIP_INDEX, ENDPOINT_NAME_TEMPLATE % ('test_type', 'cow'),
#                  {'friend__pig_cow_alliance': {'rel_type': 'friend', 'rel_key': 'pig_cow_alliance',
#                                                'source__type': 'test_type', 'source__key': 'pig',
#                                                'source__sound': 'oink', 'target__type': 'test_type',
#                                                'target__key': 'cow', 'target__sound': 'moo'}}), {})]
#
#        pig = graph_prim.Node(self.store, 'test_type', 'pig', {'sound': 'oink'})
#        cow = graph_prim.Node(self.store, 'test_type', 'cow', {'sound': 'moo'})
#        with nested(patch('pycassa.columnfamily.ColumnFamily'),
#                    patch('pycassa.batch.Mutator')) as (cf_mock, batch_mock):
#            def side_effect(self, name):
#                return name
#
#            cf_mock.side_effect = side_effect
#
#            batch_mock.return_value = BatchMutatorMock()
#            contextualized_batch_mock = batch_mock.return_value.contextualized_batch_mock
#
#            ####
#            # This is the only method tested in this case.  Crazy.
#            ####
#            pig.friend(cow, key='pig_cow_alliance')
#
#            # And, let's test our assumptions
#            self.failUnlessEqual(contextualized_batch_mock.insert.call_args_list, expected_calls)
#
#    @attr(module='agamemnon', type='unit')
#    def test_delete_relationship(self):
#        inbound_rel_index_cf = 'inbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        outbound_rel_index_cf = 'outbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        mocks = {inbound_rel_index_cf: Mock(), outbound_rel_index_cf: Mock()}
#        from_key = ENDPOINT_NAME_TEMPLATE % (type_cf, 'pig')
#        to_key = ENDPOINT_NAME_TEMPLATE % (type_cf, 'cow')
#        relationship_key = 'friend__pig_cow_alliance'
#
#        expected_calls = [((mocks[inbound_rel_index_cf], to_key), {'super_column': relationship_key}),
#            ((mocks[outbound_rel_index_cf], from_key), {'super_column': relationship_key})]
#
#        with nested(patch('pycassa.columnfamily.ColumnFamily'),
#                    patch('pycassa.batch.Mutator')) as (cf_mock, batch_mock):
#            def side_effect(self, name):
#                return mocks[name]
#
#            cf_mock.side_effect = side_effect
#
#            batch_mock.return_value = BatchMutatorMock()
#            contextualized_batch_mock = batch_mock.return_value.contextualized_batch_mock
#            rel = graph_prim.Relationship('pig_cow_alliance', graph_prim.Node(self.store, 'test_type', 'pig'),
#                                          graph_prim.Node(self.store, 'test_type', 'cow'), self.store, 'friend')
#            ####
#            # Our method under test
#            ####
#            rel.delete()
#
#            # Test our assumptions
#            self.failUnlessEqual(contextualized_batch_mock.remove.call_args_list, expected_calls)
#
#    @attr(module='agamemnon', type='unit')
#    def test_get_relationships_of_given_type_outbound_only(self):
#        inbound_rel_index_cf = 'inbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        outbound_rel_index_cf = 'outbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        out_bound_mock = Mock()
#        out_bound_mock.get.return_value = {
#            'friend__pig_cow_alliance': {'rel_type': 'friend', 'source__type': 'test_type', 'source_key': 'pig',
#                                         'target__key': 'cow', 'target__type': 'test_type',
#                                         'rel_id': 'pig_cow_alliance'}}
#        in_bound_mock = Mock()
#        in_bound_mock.get.side_effect = NotFoundException()
#        pig = graph_prim.Node(self.store, 'test_type', 'pig')
#        cow = graph_prim.Node(self.store, 'test_type', 'cow')
#
#        def side_effect(pool, type):
#            if type == outbound_rel_index_cf:
#                return out_bound_mock
#            elif type == inbound_rel_index_cf:
#                return in_bound_mock
#            else:
#                raise Exception()
#
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            cf_mock.side_effect = side_effect
#
#            ####
#            # Method under test
#            ####
#            for rel in pig.friend:
#                expected = pig.friend(cow, 'pig_cow_alliance')
#                self.failUnlessEqual(expected, rel)
#
#    @attr(module='agamemnon', type='unit')
#    def test_get_relationships_of_given_type_inbound_only(self):
#        inbound_rel_index_cf = 'inbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        outbound_rel_index_cf = 'outbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        out_bound_mock = Mock()
#        out_bound_mock.get.side_effect = NotFoundException()
#        in_bound_mock = Mock()
#        in_bound_mock.get.return_value = {
#            'friend__chicken_pig_alliance': {'rel_type': 'friend', 'source__type': 'test_type', 'source_key': 'chicken',
#                                             'target__key': 'pig', 'target__type': 'test_type',
#                                             'rel_id': 'chicken_pig_alliance'}}
#        chicken = graph_prim.Node(self.store, 'test_type', 'chicken')
#        pig = graph_prim.Node(self.store, 'test_type', 'pig')
#
#        def side_effect(pool, type):
#            if type == outbound_rel_index_cf:
#                return out_bound_mock
#            elif type == inbound_rel_index_cf:
#                return in_bound_mock
#            else:
#                raise Exception()
#
#        with patch('pycassa.columnfamily.ColumnFamily') as cf_mock:
#            cf_mock.side_effect = side_effect
#
#            ####
#            # Method under test
#            ####
#            for rel in pig.friend:
#                expected = chicken.friend(pig, 'chicken_pig_alliance')
#                self.failUnlessEqual(expected, rel)
#
#    @attr(module='agamemnon', type='unit')
#    def test_delete_node_with_relationships(self):
#        mocks = {'test_type': Mock(), 'outbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX: Mock(),
#                 'inbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX: Mock()}
#        inbound = 'inbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        outbound = 'outbound__%s' % agamemnon.graph_constants.RELATIONSHIP_INDEX
#        mocks['test_type'].remove = Mock()
#        mocks[inbound].get.side_effect = NotFoundException()
#
#        mocks[outbound].get.return_value = {
#            'friend__pig_cow_alliance': {'rel_type': 'friend', 'rel_key': 'pig_cow_alliance',
#                                         'source__type': 'test_type', 'source__key': 'pig',
#                                         'source__sound': 'oink', 'target__type': 'test_type',
#                                         'target__key': 'cow', 'target__sound': 'moo'}}
#        from_key = ENDPOINT_NAME_TEMPLATE % ('test_type', 'pig')
#        to_key = ENDPOINT_NAME_TEMPLATE % ('test_type', 'cow')
#        relationship_key = 'friend__pig_cow_alliance'
#
#        expected_calls = [((mocks[inbound], to_key), {'super_column': relationship_key}),
#            ((mocks[outbound], from_key), {}),
#            ((mocks[inbound], from_key), {}),
#            ((mocks['test_type'], 'pig'), {})]
#
#
#        def side_effect(pool, type):
#            return mocks[type]
#
#        with nested(patch('pycassa.columnfamily.ColumnFamily'),
#                    patch('pycassa.batch.Mutator')) as(cf_mock, batch_mock):
#            batch_mock.return_value = BatchMutatorMock()
#            contextual_batch_mock = batch_mock.return_value.contextualized_batch_mock
#            cf_mock.side_effect = side_effect
#            self.pig.delete()
#            call_args_list = contextual_batch_mock.remove.call_args_list
#            self.failUnlessEqual(call_args_list, expected_calls)
#
#
#



from agamemnon.graph_constants import ASCII

class InMemoryDataStore(object):
    def __init__(self):
        self.tables = {}
        self.transaction = None

    def get_cf(self, cf_name):
        if not cf_name in self.tables:
            self.tables[cf_name] = self.create_cf(cf_name)
        return self.tables[cf_name]

    def create_cf(self, type, column_type=ASCII, super=False, index_columns=list()):
        self.tables[type] = ColumnFamily(type)
        return self.tables[type]

    def cf_exists(self, type):
        return type in self.tables.keys()

    def insert(self, cf, row, columns):

        def execute():
            cf.insert(row, columns)
        if self.transactions is not None:
            self.transactions.append(execute)
        else:
            execute()

    def remove(self, cf, row, columns=None, super_column=None):
        def execute():
            cf.remove(row, columns=columns, super_column=super_column)

        if self.transactions is not None:
            self.transactions.append(execute)
        else:
            execute()

    def start_batch(self):
        self.transactions = []

    def commit_batch(self):
        for transaction in self.transactions:
            transaction()
        self.transactions = None

class ColumnFamily(object):
    def __init__(self, name):
        self.data = {}
        self.name = name

    def get(self, row, columns=None, column_start=None, super_column=None, column_finish=None, column_count=100):
        try:
            if columns is None and column_start is None and super_column is None:
                return self.data[row]
            else:

                if super_column is None:
                    columns = self.data[row]
                else:
                    columns = self.data[row][super_column]
                results = {}
                for c in columns.keys():
                    if column_start is not None:
                        if c.startswith(column_start):
                            results[c] = columns[c]
                        continue
                    results[c] = columns[c]
            return results
        except KeyError:
            return {}

    def insert(self, row, columns, ttl=None):
        if not row in self.data:
            self.data[row] = {}
        if columns is not None:
            for c in columns.keys():
                self.data[row][c] = columns[c]

            #        if ttl is not None:
            #            def delete():
            #                for c in columns.keys():
            #                    del(self.data[row][c])
            #            Timer(ttl, delete, ()).start()

    def remove(self, row, columns=None, super_column=None):
        if columns is None and super_column is None:
            del(self.data[row])
        elif super_column is None:
            row = self.data[row]
            for c in columns:
                if c in row:
                    del(row[c])
        elif columns is None:
            row = self.data[row]
            del(row[super_column])
        else:
            sc = self.data[row][super_column]
            for c in columns:
                if c in sc:
                    del(rc[c])
    def get_indexed_slices(self, index_clause):
        expression = index_clause.expressions[0]
        column_name = expression.column_name
        value = expression.value
        for i in self.data.items():
            if i[1][column_name] == value:
                yield i[0], i[1]
  